# @push.rocks/smartexpect
manage expectations in code

## Install

To install `@push.rocks/smartexpect`, use the following command in your terminal:

```bash
npm install @push.rocks/smartexpect --save
```

This will add `@push.rocks/smartexpect` to your project's dependencies. Make sure you're inside your project directory before running this command.

## Usage

`@push.rocks/smartexpect` is a TypeScript library designed to manage expectations in your code effectively, improving testing readability and maintainability. Below are various scenarios showcasing how to use this library effectively across both synchronous and asynchronous code paths.

### Getting Started

First, import `@push.rocks/smartexpect` into your TypeScript file:

```typescript
import { expect, expectAsync } from '@push.rocks/smartexpect';
```

### Synchronous Expectations

You can employ `expect` to create synchronous assertions:

```typescript
import { expect } from '@push.rocks/smartexpect';

// String type assertion
expect('hello').toBeTypeofString();

// Negated String type assertion
expect(1).not.toBeTypeofString();

// Boolean type assertion
expect(true).toBeTypeofBoolean();

// Equality assertion
expect('hithere').toEqual('hithere');

// Deep equality assertion
expect({ key: 'value' }).toEqual({ key: 'value' });

// Regular expression matching
expect('hithere').toMatch(/hi/);
```

### Asynchronous Expectations

For asynchronous operations, use `expectAsync` to return a promise:

```typescript
import { expectAsync } from '@push.rocks/smartexpect';

const asyncStringFetcher = async (): Promise<string> => {
  return 'async string';
};

const asyncTest = async () => {
  await expectAsync(asyncStringFetcher()).toBeTypeofString();
  await expectAsync(asyncStringFetcher()).toEqual('async string');
};

asyncTest();
```

### Advanced Usage

- **Properties and Deep Properties:** Assert the existence of properties and their values.
  
  ```typescript
  const testObject = { level1: { level2: 'value' } };

  // Property existence
  expect(testObject).toHaveProperty('level1');
  
  // Deep Property existence
  expect(testObject).toHaveDeepProperty(['level1', 'level2']);
  ```

- **Conditions and Comparisons:** Allow more intricate assertions like greater than, less than, or matching specific conditions.

  ```typescript
  // Greater Than
  expect(5).toBeGreaterThan(3);

  // Less Than
  expect(3).toBeLessThan(5);

  // Custom conditions
  expect(7).customAssertion(value => value > 5, 'Value is not greater than 5');
  ```

- **Arrays and Objects:** Work seamlessly with arrays and objects, checking for containment, length, or specific values.

  ```typescript
  const testArray = [1, 2, 3];

  // Containment
  expect(testArray).toContain(2);

  // Array length
  expect(testArray).toHaveLength(3);

  // Object matching
  expect({ name: 'Test', value: 123 }).toMatchObject({ name: 'Test' });
  ```

### Handling Promises and Async Operations

`@push.rocks/smartexpect` gracefully integrates with asynchronous operations, providing a `expectAsync` function that handles promise-based assertions. This keeps your tests clean and readable, irrespective of the nature of the code being tested.

### Best Practices

- **Readability:** Favor clarity and readability by explicitly stating your expectations. `@push.rocks/smartexpect`'s API is designed to be fluent and expressive, making your tests easy to write and, more importantly, easy to read.

- **Comprehensive Coverage:** Utilize the full spectrum of assertions provided to cover a broad set of use cases, ensuring your code behaves as expected not just in ideal conditions but across various edge cases.

- **Maintainability:** Group related assertions together to improve test maintainability. This makes it easier to update tests as your codebase evolves.

Through judicious use of `@push.rocks/smartexpect`, you can enhance the reliability and maintainability of your test suite, making your codebase more robust and your development workflow more efficient.

## License and Legal Information

This repository contains open-source code that is licensed under the MIT License. A copy of the MIT License can be found in the [license](license) file within this repository. 

**Please note:** The MIT License does not grant permission to use the trade names, trademarks, service marks, or product names of the project, except as required for reasonable and customary use in describing the origin of the work and reproducing the content of the NOTICE file.

### Trademarks

This project is owned and maintained by Task Venture Capital GmbH. The names and logos associated with Task Venture Capital GmbH and any related products or services are trademarks of Task Venture Capital GmbH and are not included within the scope of the MIT license granted herein. Use of these trademarks must comply with Task Venture Capital GmbH's Trademark Guidelines, and any usage must be approved in writing by Task Venture Capital GmbH.

### Company Information

Task Venture Capital GmbH  
Registered at District court Bremen HRB 35230 HB, Germany

For any legal inquiries or if you require further information, please contact us via email at hello@task.vc.

By using this repository, you acknowledge that you have read this section, agree to comply with its terms, and understand that the licensing of the code does not imply endorsement by Task Venture Capital GmbH of any derivative works.
